@@ -31,6 +31,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <errno.h>
 
 #include <glib.h>
 #include <gmodule.h>
@@ -70,6 +71,7 @@ struct str_enchant_broker
 	GSList *provider_list;	/* list of all of the spelling backend providers */
 	GHashTable *dict_map;		/* map of language tag -> dictionary */
 	GHashTable *provider_ordering; /* map of language tag -> provider order */
+	GHashTable *params;
 
 	gchar * error;
 };
@@ -104,6 +106,33 @@ typedef void             (*EnchantPreConfigureFunc) (EnchantProvider * provider,
 /********************************************************************************/
 /********************************************************************************/
 
+
+static GSList* enchant_slist_prepend_unique_string (GSList *slist, gchar* data)
+{
+	if (NULL == g_slist_find_custom (slist, data, (GCompareFunc)strcmp))
+		{
+			return g_slist_prepend (slist, data);
+		}
+	else
+		{
+			g_free (data);
+			return slist;
+		}
+}
+
+static GSList* enchant_slist_append_unique_string (GSList *slist, gchar* data)
+{
+	if (NULL == g_slist_find_custom (slist, data, (GCompareFunc)strcmp))
+		{
+			return g_slist_append (slist, data);
+		}
+	else
+		{
+			g_free (data);
+			return slist;
+		}
+}
+
 static GSList *
 _enchant_get_user_home_dirs (void)
 {
@@ -112,25 +141,22 @@ _enchant_get_user_home_dirs (void)
 
 	home_dir = enchant_get_registry_value_ex (1, "Config", "Home_Dir");
 	if (home_dir)
-		dirs = g_slist_append (dirs, g_strdup (home_dir));
+		dirs = enchant_slist_append_unique_string (dirs, g_strdup (home_dir));
 
 	home_dir = g_get_home_dir ();
 	if (home_dir)
-		dirs = g_slist_append (dirs, g_strdup (home_dir));
+		dirs = enchant_slist_append_unique_string (dirs, g_strdup (home_dir));
 
 	return dirs;
 }
 
 static void
-_enchant_ensure_private_datadir (void)
+_enchant_ensure_dir_exists (const char* dir)
 {
-	const char * config_dir;
-
-	config_dir = g_get_user_config_dir();
-	if (config_dir && !g_file_test (config_dir, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) 
+	if (dir && !g_file_test (dir, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) 
 		{
-			(void)g_remove (config_dir);
-			g_mkdir_with_parents (config_dir, 0700);                        
+			(void)g_remove (dir);
+			g_mkdir_with_parents (dir, 0700);                        
 		}
 }
 
@@ -145,7 +171,7 @@ enchant_get_user_dirs (void)
 		user_config_dir = g_get_user_config_dir();
 		
 		if (user_config_dir)
-			user_dirs = g_slist_append (user_dirs, g_build_filename (user_config_dir,
+			user_dirs = enchant_slist_append_unique_string (user_dirs, g_build_filename (user_config_dir,
 										 "enchant",
 										 NULL));
 	}
@@ -156,7 +182,7 @@ enchant_get_user_dirs (void)
 		
 		for (dir = home_dirs; dir; dir = dir->next)
 			{
-				user_dirs = g_slist_append (user_dirs,
+				user_dirs = enchant_slist_append_unique_string (user_dirs,
 							    g_build_filename (dir->data,
 									      ENCHANT_USER_PATH_EXTENSION,
 									      NULL));
@@ -183,11 +209,11 @@ enchant_get_module_dirs (void)
 		
 		user_module_dir = enchant_get_registry_value_ex (1, "Config", "Module_Dir");
 		if (user_module_dir)
-			module_dirs = g_slist_append (module_dirs, user_module_dir);
+			module_dirs = enchant_slist_append_unique_string (module_dirs, user_module_dir);
 	}
 
 #ifdef XP_TARGET_COCOA
-	module_dirs = g_slist_append (module_dirs, g_strdup ([[EnchantResourceProvider instance] moduleFolder]));
+	module_dirs = enchant_slist_append_unique_string (module_dirs, g_strdup ([[EnchantResourceProvider instance] moduleFolder]));
 #endif
 
 	{
@@ -196,7 +222,7 @@ enchant_get_module_dirs (void)
 		user_dirs = enchant_get_user_dirs();
 
 		for (iter = user_dirs; iter; iter = iter->next)
-			module_dirs = g_slist_append (module_dirs, iter->data);
+			module_dirs = enchant_slist_append_unique_string (module_dirs, iter->data);
 		
 		g_slist_free (user_dirs);
 	}
@@ -204,7 +230,7 @@ enchant_get_module_dirs (void)
 	/* Look for explicitly set registry values */
 	module_dir = enchant_get_registry_value_ex (0, "Config", "Module_Dir");
 	if (module_dir)
-		module_dirs = g_slist_append (module_dirs, module_dir);
+		module_dirs = enchant_slist_append_unique_string (module_dirs, module_dir);
 
 	/* Dynamically locate library and search for modules relative to it. */
 	prefix = enchant_get_prefix_dir();
@@ -212,11 +238,11 @@ enchant_get_module_dirs (void)
 		{
 			module_dir = g_build_filename(prefix,"lib","enchant",NULL);
 			g_free(prefix);
-			module_dirs = g_slist_append (module_dirs, module_dir);
+			module_dirs = enchant_slist_append_unique_string (module_dirs, module_dir);
 		}
 
 #if defined(ENCHANT_GLOBAL_MODULE_DIR)
-	module_dirs = g_slist_append (module_dirs, g_strdup (ENCHANT_GLOBAL_MODULE_DIR));
+	module_dirs = enchant_slist_append_unique_string (module_dirs, g_strdup (ENCHANT_GLOBAL_MODULE_DIR));
 #endif
 
 	return module_dirs;
@@ -225,17 +251,26 @@ enchant_get_module_dirs (void)
 static GSList *
 enchant_get_conf_dirs (void)
 {
-	GSList *conf_dirs = NULL;
-	char * ordering_dir = NULL, * prefix = NULL;;
+	GSList *conf_dirs = NULL, *user_conf_dirs, *iter;
+	char * ordering_dir = NULL, * prefix = NULL;
+
+	user_conf_dirs = enchant_get_user_config_dirs();
+
+	for (iter = user_conf_dirs; iter != NULL; iter = iter->next)
+		{
+			conf_dirs = enchant_slist_append_unique_string (conf_dirs, iter->data);
+		}
+
+	g_slist_free (user_conf_dirs);
 
 #ifdef XP_TARGET_COCOA
-	conf_dirs = g_slist_append (conf_dirs, g_strdup ([[EnchantResourceProvider instance] configFolder]));
+	conf_dirs = enchant_slist_append_unique_string (conf_dirs, g_strdup ([[EnchantResourceProvider instance] configFolder]));
 #endif
 
 	/* Look for explicitly set registry values */
 	ordering_dir = enchant_get_registry_value_ex (0, "Config", "Data_Dir");
 	if (ordering_dir)
-		conf_dirs = g_slist_append (conf_dirs, ordering_dir);
+		conf_dirs = enchant_slist_append_unique_string (conf_dirs, ordering_dir);
 
 	/* Dynamically locate library and search for files relative to it. */
 	prefix = enchant_get_prefix_dir();
@@ -243,16 +278,53 @@ enchant_get_conf_dirs (void)
 		{
 			ordering_dir = g_build_filename(prefix,"share","enchant",NULL);
 			g_free(prefix);
-			conf_dirs = g_slist_append (conf_dirs, ordering_dir);
+			conf_dirs = enchant_slist_append_unique_string (conf_dirs, ordering_dir);
 		}
 
 #if defined(ENCHANT_GLOBAL_ORDERING)
-	conf_dirs = g_slist_append (conf_dirs, g_strdup (ENCHANT_GLOBAL_ORDERING));
+	conf_dirs = enchant_slist_append_unique_string (conf_dirs, g_strdup (ENCHANT_GLOBAL_ORDERING));
 #endif
 
 	return conf_dirs;
 }
 
+ENCHANT_MODULE_EXPORT(FILE *)
+enchant_fopen (const gchar *filename, const gchar *mode)
+{
+#ifdef G_OS_WIN32
+  wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
+  wchar_t *wmode;
+  FILE *retval;
+  int save_errno;
+
+  if (wfilename == NULL)
+    {
+      errno = EINVAL;
+      return NULL;
+    }
+
+  wmode = g_utf8_to_utf16 (mode, -1, NULL, NULL, NULL);
+
+  if (wmode == NULL)
+    {
+      g_free (wfilename);
+      errno = EINVAL;
+      return NULL;
+    }
+
+  retval = _wfopen (wfilename, wmode);
+  save_errno = errno;
+
+  g_free (wfilename);
+  g_free (wmode);
+
+  errno = save_errno;
+  return retval;
+#else
+  return fopen (filename, mode);
+#endif
+}
+
 /**
  * enchant_get_user_config_dir
  *
@@ -274,7 +346,7 @@ enchant_get_user_config_dirs (void)
 
 	user_config = enchant_get_registry_value_ex (1, "Config", "Data_Dir");
 	if (user_config)
-		dirs = g_slist_prepend (dirs, user_config);
+		dirs = enchant_slist_prepend_unique_string (dirs, user_config);
 
 	return dirs;
 }
@@ -308,13 +380,13 @@ enchant_get_registry_value_ex (int current_user, const char * const prefix, cons
 	uKeyName = g_utf8_to_utf16 (keyName, -1, NULL, NULL, NULL);
 	uKey = g_utf8_to_utf16 (key, -1, NULL, NULL, NULL);
 
-	if(RegOpenKeyEx(baseKey, uKeyName, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
+	if(RegOpenKeyExW(baseKey, uKeyName, 0, KEY_READ, &hKey) == ERROR_SUCCESS)
 		{
 			/* Determine size of string */
-			if(RegQueryValueEx( hKey, uKey, NULL, &lType, NULL, &dwSize) == ERROR_SUCCESS)
+			if(RegQueryValueExW( hKey, uKey, NULL, &lType, NULL, &dwSize) == ERROR_SUCCESS)
 				{
 					wszValue = g_new0(WCHAR, dwSize + 1);
-					RegQueryValueEx(hKey, uKey, NULL, &lType, (LPBYTE) wszValue, &dwSize);
+					RegQueryValueExW(hKey, uKey, NULL, &lType, (LPBYTE) wszValue, &dwSize);
 				}
 		}
 
@@ -530,6 +602,7 @@ _enchant_session_new (EnchantProvider *provider, const char * const user_config_
 	session = enchant_session_new_with_pwl (provider, dic, excl, lang, fail_if_no_pwl);
 	
 	g_free (dic);
+	g_free (excl);
 
 	return session;
 }
@@ -540,21 +613,22 @@ enchant_session_new (EnchantProvider *provider, const char * const lang)
 	EnchantSession * session = NULL;
 	GSList *user_config_dirs, *iter;
 
-	_enchant_ensure_private_datadir ();
-
 	user_config_dirs = enchant_get_user_config_dirs ();
 	for (iter = user_config_dirs; iter != NULL && session == NULL; iter = iter->next)
 		{
 			session =_enchant_session_new (provider, iter->data, lang, TRUE);
 		}
 
+	if (session == NULL && user_config_dirs != NULL)
+		{
+			_enchant_ensure_dir_exists (user_config_dirs->data);
+	
+			session =_enchant_session_new (provider, user_config_dirs->data, lang, FALSE);
+		}
+
 	g_slist_foreach (user_config_dirs, (GFunc)g_free, NULL);
 	g_slist_free (user_config_dirs);
 
-	if (!session)
-		{
-			session =_enchant_session_new (provider, g_get_user_config_dir(), lang, FALSE);
-		}
 	
 	return session;
 }
@@ -1435,7 +1509,7 @@ enchant_load_ordering_from_file (EnchantBroker * broker, const char * file)
 
 	FILE * f;
 
-	f = g_fopen (file, "r");
+	f = enchant_fopen (file, "r");
 	if (!f)
 		return;
 
@@ -1598,7 +1672,7 @@ enchant_broker_init (void)
 	
 	broker->dict_map = g_hash_table_new_full (g_str_hash, g_str_equal,
 						  g_free, enchant_dict_destroyed);
-	
+	broker->params = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
 	enchant_load_providers (broker);	
 	enchant_load_provider_ordering (broker);
 
@@ -1627,6 +1701,7 @@ enchant_broker_free (EnchantBroker * broker)
 	/* will destroy any remaining dictionaries for us */
 	g_hash_table_destroy (broker->dict_map);
 	g_hash_table_destroy (broker->provider_ordering);
+	g_hash_table_destroy (broker->params);
 	
 	g_slist_foreach (broker->provider_list, enchant_provider_free, NULL);
 	g_slist_free (broker->provider_list);
@@ -1694,6 +1769,7 @@ _enchant_broker_request_dict (EnchantBroker * broker, const char *const tag)
 {
 	EnchantDict * dict;
 	GSList * list;
+	GSList * listIter;
 
 	dict = (EnchantDict*)g_hash_table_lookup (broker->dict_map, (gpointer) tag);
 	if (dict) {
@@ -1701,11 +1777,12 @@ _enchant_broker_request_dict (EnchantBroker * broker, const char *const tag)
 		return dict;
 	}
 
-	for (list = enchant_get_ordered_providers (broker, tag); list != NULL; list = g_slist_next (list))
+	list = enchant_get_ordered_providers (broker, tag);
+	for (listIter = list; listIter != NULL; listIter = g_slist_next (listIter))
 		{
 			EnchantProvider * provider;
 
-			provider = (EnchantProvider *) list->data;
+			provider = (EnchantProvider *) listIter->data;
 			
 			if (provider->request_dict)
 				{
@@ -2153,17 +2230,18 @@ enchant_get_prefix_dir(void)
 #ifdef _WIN32
 	if (!prefix) {
 		/* Dynamically locate library and return containing directory */
-		HINSTANCE hInstance = GetModuleHandle(L"libenchant");
-		if(hInstance != NULL)
+		WCHAR dll_path[MAX_PATH];
+
+		if(GetModuleFileNameW(s_hModule,dll_path,MAX_PATH))
 			{
-				WCHAR dll_path[MAX_PATH];
-	  
-				if(GetModuleFileName(hInstance,dll_path,MAX_PATH))
-					{
-						gchar* utf8_dll_path = g_utf16_to_utf8 (dll_path, -1, NULL, NULL, NULL);
-						prefix = g_path_get_dirname(utf8_dll_path);
-						g_free(utf8_dll_path);
-					}
+				gchar* utf8_dll_path = g_utf16_to_utf8 (dll_path, -1, NULL, NULL, NULL);
+				prefix = g_path_get_dirname(utf8_dll_path);
+				g_free(utf8_dll_path);
+				/* Strip off "bin" subfolder if present */
+				if (strlen(prefix) >=6 &&
+				    G_IS_DIR_SEPARATOR(prefix[strlen(prefix)-4]) &&
+				    g_ascii_strcasecmp(prefix+strlen(prefix)-3, "bin") == 0)
+					prefix[strlen(prefix)-4] = '\0';
 			}
 	}
 #endif
@@ -2183,3 +2261,54 @@ enchant_get_prefix_dir(void)
 
 	return prefix;
 }
+
+ENCHANT_MODULE_EXPORT(char *)
+enchant_broker_get_param (EnchantBroker * broker, const char * const param_name) 
+{
+	g_return_val_if_fail (broker, NULL);
+	g_return_val_if_fail (param_name && *param_name, NULL);
+
+	return g_hash_table_lookup (broker->params, param_name);
+}
+
+ENCHANT_MODULE_EXPORT(void)
+enchant_broker_set_param (EnchantBroker * broker, const char * const param_name, const char * const param_value)
+{
+	g_return_if_fail (broker);
+	g_return_if_fail (param_name && *param_name);
+
+	if (param_value == NULL || *param_value == '\0')
+		g_hash_table_remove (broker->params, param_name);
+	else
+		g_hash_table_insert (broker->params, g_strdup (param_name), g_strdup (param_value));
+}
+
+ENCHANT_MODULE_EXPORT (GSList *)
+enchant_get_dirs_from_param (EnchantBroker * broker, const char * const param_name)
+{
+	const char *param_value;
+	char **tokens;
+	GSList *dirs = NULL;
+
+	param_value = enchant_broker_get_param (broker, param_name);
+	if (param_value == NULL)
+		return NULL;
+
+#ifdef _WIN32
+	tokens = g_strsplit (param_value, ";", 0);
+#else
+	tokens = g_strsplit (param_value, ":", 0);
+#endif
+	if (tokens != NULL) {
+		int i;
+		for (i = 0; tokens[i]; i++) 
+			{
+				char *token = g_strstrip(tokens[i]);
+				dirs = g_slist_append (dirs, g_strdup (token));
+			}
+		
+		g_strfreev (tokens);		
+	}
+
+	return dirs;
+}
